##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import numpy as np
from pyquil.paulis import PauliTerm, PauliSum
import pyquil.api as api
from scipy.optimize import minimize
from grove.pyqaoa.qaoa import QAOA
CXN = api.SyncConnection()


def numpart_qaoa(asset_list, A=1.0, minimizer_kwargs=None, steps=1):
    """
    generate number partition driver and cost functions

    :param asset_list: list to binary parition
    :param A: (float) optional constant for level separation. Default=1.
    :param minimizer_kwargs: Arguments for the QAOA minimizer
    :param steps: (int) number of steps approximating the solution.
    """
    cost_operators = []
    ref_operators = []
    for ii in range(len(asset_list)):
        for jj in range(ii + 1, len(asset_list)):
            cost_operators.append(PauliSum([PauliTerm("Z", ii, 2*asset_list[ii]) *
                                            PauliTerm("Z", jj, A*asset_list[jj])]))
        ref_operators.append(PauliSum([PauliTerm("X", ii, -1.0)]))

    cost_operators.append(PauliSum([PauliTerm("I", 0, len(asset_list))]))

    if minimizer_kwargs is None:
        minimizer_kwargs = {'method': 'Nelder-Mead',
                            'options': {'ftol': 1.0e-2,
                                        'xtol': 1.0e-2,
                                        'disp': True}}
    n_qubits = len(asset_list)
    qaoa_inst = QAOA(CXN, n_qubits, steps=steps, cost_ham=cost_operators,
                     ref_hamiltonian=ref_operators, store_basis=True,
                     minimizer=minimize, minimizer_kwargs=minimizer_kwargs,
                     vqe_options={'disp': True})

    return qaoa_inst


if __name__ == "__main__":
    # Sample Run.
    # result should be an even partition of nodes
    inst = numpart_qaoa([1, 1, 1, 1, 1, 1], A=1.0, steps=1)
    betas, gammas = inst.get_angles()
    print(betas)
    print(gammas)
    probs = inst.probabilities(np.hstack((betas, gammas)))
    for state, prob in zip(inst.states, probs):
        print(state, prob)

    print("Most frequent bitstring from sampling")
    most_freq_string, sampling_results = inst.get_string(
            betas, gammas, samples=100)
    print(most_freq_string)



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Module for Grover's algorithm.
"""

import numpy as np
import pyquil.quil as pq
from pyquil.gates import H
from pyquil.quilbase import Qubit

from grove.amplification.amplification import amplification_circuit


class Grover(object):
    """This class contains an implementation of Grover's algorithm using pyQuil. See `these notes`_
     by Dave Bacon for more information.

    .. _these notes: https://courses.cs.washington.edu/courses/cse599d/06wi/lecturenotes12.pdf
    """
    def __init__(self):
        self.unitary_function_mapping = None
        self.n_qubits = None
        self.qubits = None
        self.grover_circuit = None
        self.bit_map = None

    @staticmethod
    def _compute_grover_oracle_matrix(bitstring_map):
        """Computes the unitary matrix that encodes the oracle function for Grover's algorithm

        :param bitstring_map: dict with string keys corresponding to bitstrings,
         and integer values corresponding to the desired phase on the output state.
        :type bitstring_map: Dict[String, Int]
        :return: a numpy array corresponding to the unitary matrix for oracle for the given
         bitstring_map
        :rtype: numpy.ndarray
        """
        n_bits = len(list(bitstring_map.keys())[0])
        oracle_matrix = np.zeros(shape=(2 ** n_bits, 2 ** n_bits))
        for b in range(2 ** n_bits):
            pad_str = np.binary_repr(b, n_bits)
            phase_factor = bitstring_map[pad_str]
            oracle_matrix[b, b] = phase_factor
        return oracle_matrix

    def _construct_grover_circuit(self):
        """Contructs an instance of Grover's Algorithm, using initialized values.

        :return: None
        :rtype: NoneType
        """
        oracle = pq.Program()
        oracle_name = "GROVER_ORACLE"
        oracle.defgate(oracle_name, self.unitary_function_mapping)
        oracle.inst(tuple([oracle_name] + self.qubits))
        self.grover_circuit = self.oracle_grover(oracle, self.qubits)

    def _init_attr(self, bitstring_map):
        """Initializes an instance of Grover's Algorithm given a bitstring_map.

        :param bitstring_map: dict with string keys corresponding to bitstrings, and integer
         values corresponding to the desired phase on the output state.
        :type bitstring_map: Dict[String, Int]
        :return: None
        :rtype: NoneType
        """
        self.bit_map = bitstring_map
        self.unitary_function_mapping = self._compute_grover_oracle_matrix(bitstring_map)
        self.n_qubits = self.unitary_function_mapping.shape[0]
        self.qubits = list(range(int(np.log2(self.n_qubits))))
        self._construct_grover_circuit()

    def find_bitstring(self, cxn, bitstring_map):
        """
        Runs Grover's Algorithm to find the bitstring that is designated by ``bistring_map``.

        In particular, this will prepare an initial state in the uniform superposition over all bit-
        strings, an then use Grover's Algorithm to pick out the desired bitstring.

        :param JobConnection cxn: the connection to the Rigetti cloud to run pyQuil programs.
        :param bitstring_map: a mapping from bitstrings to the phases that the oracle should impart
            on them. If the oracle should "look" for a bitstring, it should have a ``-1``, otherwise
            it should have a ``1``.
        :type bitstring_map: Dict[String, Int]
        :return: Returns the bitstring resulting from measurement after Grover's Algorithm.
        :rtype: str
        """
        self._init_attr(bitstring_map)
        sampled_bitstring = cxn.run_and_measure(self.grover_circuit, self.qubits)[0]
        return ''.join([str(b) for b in sampled_bitstring])

    @staticmethod
    def oracle_grover(oracle, qubits, num_iter=None):
        r"""Implementation of Grover's Algorithm for a given oracle.

        :param Program oracle: An oracle defined as a Program. It should send :math:`\ket{x}`
            to :math:`(-1)^{f(x)}\ket{x}`, where the range of f is {0, 1}.
        :param qubits: List of qubits for Grover's Algorithm.
        :type qubits: list[int or Qubit]
        :param int num_iter: The number of iterations to repeat the algorithm for.
                             The default is the integer closest to :math:`\frac{\pi}{4}\sqrt{N}`,
                             where :math:`N` is the size of the domain.
        :return: A program corresponding to the desired instance of Grover's Algorithm.
        :rtype: Program
        """
        if num_iter is None:
            num_iter = int(round(np.pi * 2 ** (len(qubits) / 2.0 - 2.0)))
        uniform_superimposer = pq.Program().inst([H(qubit) for qubit in qubits])
        amp_prog = amplification_circuit(uniform_superimposer, oracle, qubits, num_iter)
        return amp_prog



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""
Module for  Simon's Algorithm.
For more information, see [Simon1995]_, [Loceff2015]_, [Watrous2006]_

.. [Simon1995] Simon, D.R. (1995), `"On the power of quantum computation"`_,
 35th Annual Symposium on Foundations of Computer Science, Proceedings, p. 116-123.

.. [Loceff2015] Loceff, M. (2015), `"A Course in Quantum Computing for the Community College"`_,
 Volume 1, Chapter 18, p 484-541.

.. [Watrous2006] Watrous, J. (2006), `"Simon's Algorithm"`_, University of Calgary CPSC 519/619:
 Quantum Computation, Lecture 6.

.. _`"On the power of quantum computation"`: https://courses.cs.washington.edu/courses/cse599/01wi/
 papers/simon_qc.pdf

.. _`"A Course in Quantum Computing for the Community College"`: http://lapastillaroja.net/
 wp-content/uploads/2016/09/Intro_to_QC_Vol_1_Loceff.pdf

.. _`"Simon's Algorithm"`: https://cs.uwaterloo.ca/~watrous/CPSC519/LectureNotes/06.pdf
"""

from collections import defaultdict
from operator import xor

import numpy as np
import numpy.random as rd
import pyquil.quil as pq
from pyquil.gates import H

import grove.simon.utils as utils


def create_1to1_bitmap(mask):
    """
    A helper to create a bit map function (as a dictionary) for a given mask. E.g. for a mask
    :math:`m = 10` the return is a dictionary:

    >>> create_1to1_bitmap('10')
    ... {
    ...     '00': '10',
    ...     '01': '11',
    ...     '10': '00',
    ...     '11': '01'
    ... }

    :param String mask: binary mask as a string of 0's and 1's
    :return: dictionary containing a mapping of all possible bit strings of the same length as the
        mask's string and their mapped bit-string value
    :rtype: Dict[String, String]
    """
    n_bits = len(mask)
    form_string = "{0:0" + str(n_bits) + "b}"
    bit_map_dct = {}
    for idx in range(2**n_bits):
        bit_string = form_string.format(idx)
        bit_map_dct[bit_string] = utils.bitwise_xor(bit_string, mask)
    return bit_map_dct


def create_valid_2to1_bitmap(mask, random_seed=None):
    """
    A helper to create a 2-to-1 binary function that is invariant with respect to the application of
    a specified XOR bitmask. This property must be satisfied if a 2-to-1 function is to be used in
    Simon's algorithm

    More explicitly, such a 2-to-1 function :math:`f` must satisfy :math:`f(x) = f(x \oplus m)`
    where :math:`m` is a bit mask and :math:`\oplus` denotes the bit wise XOR operation. An example
    of such a function is the truth-table

    ====  ====
     x    f(x)
    ====  ====
    000   101
    001   010
    010   000
    011   110
    100   000
    101   110
    110   101
    111   010
    ====  ====

    Note that, e.g. both `000` and `110` map to the same value `101` and
    :math:`000 \oplus 110 = 110`. The same holds true for other pairs.

    :param String mask: mask input that defines the periodicity of f
    :param Integer random_seed: (optional) integer to set numpy.random.seed parameter.
    :return: dictionary containing the truth table of a valid 2-to-1 boolean function
    :rtype: Dict[String, String]
    """
    if random_seed is not None:
        rd.seed(random_seed)
    bit_map = create_1to1_bitmap(mask)
    n_samples = int(len(bit_map.keys()) / 2)

    # We create a 2-to-1 mapping and hence need to generate a list with exactly half the possible
    # bit-strings. We do this by randomly sampling the set of all possible bit-strings.
    # Moreover we sort the keys as the return order is not guaranteed and we want to remove
    # this randomness
    range_of_2to1_map = list(rd.choice(list(sorted(bit_map.keys())), replace=False, size=n_samples))

    list_of_bitstring_tuples = sorted([(k, v) for k, v in bit_map.items()], key=lambda x: x[0])

    bit_map_dct = {}
    for cnt in range(n_samples):
        bitstring_tup = list_of_bitstring_tuples[cnt]
        val = range_of_2to1_map[cnt]
        bit_map_dct[bitstring_tup[0]] = val
        bit_map_dct[bitstring_tup[1]] = val

    return bit_map_dct


class Simon(object):
    """
    This class contains an implementation of Simon's algorithm using pyQuil. For more references see
    the documentation_

    .. _documentation: http://grove-docs.readthedocs.io/en/latest/simon.html
    """

    def __init__(self):
        self.unitary_function_mapping = None
        self.n_qubits = None
        self.n_ancillas = None
        self._qubits = None
        self.computational_qubits = None
        self.ancillas = None
        self.simon_circuit = None
        self._dict_of_linearly_indep_bit_vectors = {}
        self.mask = None
        self.bit_map = None
        self.classical_register = None

    def _construct_simon_circuit(self):
        """
        Implementation of the quantum portion of Simon's Algorithm.

        Given a list of input qubits, all initially in the :math:`\\vert 0\\rangle` state, create a
        program that applies the Walsh-Hadamard transform the qubits before and after going through
        the oracle.

        :return: A program corresponding to the desired instance of Simon's Algorithm.
        :rtype: Program
        """
        simon_circuit = pq.Program()

        oracle_name = "SIMON_ORACLE"
        simon_circuit.defgate(oracle_name, self.unitary_function_mapping)

        simon_circuit.inst([H(i) for i in self.computational_qubits])
        simon_circuit.inst(tuple([oracle_name] + sorted(self._qubits, reverse=True)))
        simon_circuit.inst([H(i) for i in self.computational_qubits])
        return simon_circuit

    def _init_attr(self, bitstring_map):
        """
        Acts instead of __init__ method to instantiate the necessary Simon Object state.

        :param Dict[String, String] bitstring_map: truth-table of the input bitstring map in
        dictionary format
        :return: None
        :rtype: NoneType
        """
        self.bit_map = bitstring_map
        self.n_qubits = len(list(bitstring_map.keys())[0])
        self.n_ancillas = self.n_qubits
        self._qubits = list(range(self.n_qubits + self.n_ancillas))
        self.computational_qubits = self._qubits[:self.n_qubits]
        self.ancillas = self._qubits[self.n_qubits:]
        self.unitary_function_mapping, _ = self._compute_unitary_oracle_matrix(bitstring_map)
        self.simon_circuit = self._construct_simon_circuit()
        self._dict_of_linearly_indep_bit_vectors = {}
        self.mask = None

    @staticmethod
    def _compute_unitary_oracle_matrix(bitstring_map):
        """
        Computes the unitary matrix that encodes the orcale function for Simon's algorithm

        :param Dict[String, String] bitstring_map: truth-table of the input bitstring map in
        dictionary format
        :return: a dense matrix containing the permutation of the bit strings and a dictionary
        containing the indices of the non-zero elements of the computed permutation matrix as
        key-value-pairs
        :rtype: Tuple[2darray, Dict[String, String]]
        """
        n_bits = len(list(bitstring_map.keys())[0])

        # We instantiate an empty matrix of size 2 * n_bits to encode the mapping from n qubits
        # to n ancillas, which explains the factor 2 overhead.
        # To construct the matrix we go through all possible state transitions and pad the index
        # according to all possible states the ancilla-subsystem could be in
        ufunc = np.zeros(shape=(2 ** (2 * n_bits), 2 ** (2 * n_bits)))
        index_mapping_dct = defaultdict(dict)
        for b in range(2**n_bits):
            # padding according to ancilla state
            pad_str = np.binary_repr(b, n_bits)
            for k, v in bitstring_map.items():
                # add mapping from initial state to the state in the ancilla system.
                # pad_str corresponds to the initial state of the ancilla system.
                index_mapping_dct[pad_str + k] = utils.bitwise_xor(pad_str, v) + k
                # calculate matrix indices that correspond to the transition-matrix-element
                # of the oracle unitary
                i, j = int(pad_str+k, 2), int(utils.bitwise_xor(pad_str, v) + k, 2)
                ufunc[i, j] = 1
        return ufunc, index_mapping_dct

    def find_mask(self, cxn, bitstring_map):
        """
        Runs Simon's mask_array algorithm to find the mask.

        :param JobConnection cxn: the connection to the Rigetti cloud to run pyQuil programs
        :param Dict[String, String] bitstring_map: a truth table describing the boolean function,
            whose period is  to be found.

        :return: Returns the mask of the bitstring map or raises an Exception if the mask cannot be
            found.
        :rtype: String
        """
        self._init_attr(bitstring_map)

        # create the samples of linearly independent bit-vectors
        self._sample_independent_bit_vectors(cxn)
        # try to invert the mask and check validity
        self._invert_mask_equation()

        if self._check_mask_correct():
            return self.mask
        else:
            raise Exception("No valid mask found")

    def _sample_independent_bit_vectors(self, cxn):
        """This method samples :math:`n-1` linearly independent vectors using the Simon Circuit.
        It attempts to put the sampled bitstring into a dictionary and only terminates once the
        dictionary contains :math:`n-1` samples

        :param cxn: Connection object to a QVM or QPU
        :return: None
        :rtype: NoneType
        """
        while len(self._dict_of_linearly_indep_bit_vectors) < self.n_qubits - 1:
            sampled_bit_string = np.array(cxn.run_and_measure(self.simon_circuit,
                                                              self.computational_qubits)[0],
                                          dtype=int)
            self._add_to_dict_of_indep_bit_vectors(sampled_bit_string.tolist())

    def _invert_mask_equation(self):
        """
        This method tries to infer the bit mask of the input function from the sampled :math:`n-1`
        linearly independent bit vectors.

        It first finds the value of the missing most-significant bit (MSB) in the collection of
        sampled bit vectors, then constructs a matrix in upper-triangular (row-echelon) form and
        finally uses back-substitution over :math:`GF(2)` to find a solution to the equation

            :math:`\\mathbf{\\mathit{W}}\\mathbf{m}=\\mathbf{a}`

        where :math:`a` represents the bit vector of missing provenance, :math:`\mathbf{m}` is the
        mask to be found and :math:`\\mathbf{\\mathit{W}}` is the constructed upper-triangular
        matrix.

        :return: None
        :rtype: NoneType
        """
        missing_msb = self._add_missing_msb_vector()
        upper_triangular_matrix = np.asarray(
            [tup[1] for tup in sorted(zip(self._dict_of_linearly_indep_bit_vectors.keys(),
                                          self._dict_of_linearly_indep_bit_vectors.values()),
                                      key=lambda x: x[0])])

        msb_unit_vec = np.zeros(shape=(self.n_qubits,), dtype=int)
        msb_unit_vec[missing_msb] = 1

        self.mask = utils.binary_back_substitute(upper_triangular_matrix, msb_unit_vec).tolist()

    def _add_to_dict_of_indep_bit_vectors(self, z):
        """
        This method adds a bit-vector z to the dictionary of independent vectors. It checks the
        provenance (most significant bit) of the vector and only adds it to the dictionary if the
        provenance is not yet found in the dictionary. This guarantees that we can write up a
        resulting matrix in upper-triangular form which by virtue of its form is invertible

        :param z: array containing the bit-vector
        :return: None
        :rtype: NoneType
        """
        if all(np.asarray(z) == 0) or all(np.asarray(z) == 1):
            return None
        msb_z = utils.most_significant_bit(z)

        # try to add bitstring z to samples dictionary directly
        if msb_z not in self._dict_of_linearly_indep_bit_vectors.keys():
            self._dict_of_linearly_indep_bit_vectors[msb_z] = z
        # if we have a conflict with the provenance of a sample try to create
        # bit-wise XOR vector (guaranteed to be orthogonal to the conflict) and add
        # that to the samples.
        # Bail if this doesn't work and continue sampling.
        else:
            conflict_z = self._dict_of_linearly_indep_bit_vectors[msb_z]
            not_z = [xor(conflict_z[idx], z[idx]) for idx in range(len(z))]
            if all(np.asarray(not_z) == 0):
                return None
            msb_not_z = utils.most_significant_bit(not_z)
            if msb_not_z not in self._dict_of_linearly_indep_bit_vectors.keys():
                self._dict_of_linearly_indep_bit_vectors[msb_not_z] = not_z

    def _add_missing_msb_vector(self):
        """
        Finds the missing provenance value in the collection of :math:`n-1` linearly independent
        bit vectors and adds a unit vector corresponding to the missing provenance to the collection

        :return: Missing provenance value as int
        :rtype: Int
        """
        missing_msb = None
        for idx in range(self.n_qubits):
            if idx not in self._dict_of_linearly_indep_bit_vectors.keys():
                missing_msb = idx

        if missing_msb is None:
            raise ValueError("Expected a missing provenance, but didn't find one.")

        augment_vec = np.zeros(shape=(self.n_qubits,))
        augment_vec[missing_msb] = 1
        self._dict_of_linearly_indep_bit_vectors[missing_msb] = augment_vec.astype(int).tolist()
        return missing_msb

    def _check_mask_correct(self):
        """
        Checks if a given mask correctly reproduces the function that was provided to the Simon
        algorithm. This can be done in :math:`O(n)` as it is a simple list traversal.

        :return: True if mask reproduces the input function
        :rtype: Bool
        """
        mask_str = ''.join([str(b) for b in self.mask])
        return all([self.bit_map[k] == self.bit_map[utils.bitwise_xor(k, mask_str)]
                    for k in self.bit_map.keys()])



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

import pyquil.quil as pq
from pyquil.gates import SWAP, H, CPHASE
import math


def bit_reversal(qubits):
    """
    Generate a circuit to do bit reversal.

    :param qubits: Qubits to do bit reversal with.
    :return: A program to do bit reversal.
    """
    p = pq.Program()
    n = len(qubits)
    for i in range(int(n / 2)):
        p.inst(SWAP(qubits[i], qubits[-i - 1]))
    return p


def _core_qft(qubits, coeff):
    """
    Generates the core program to perform the quantum Fourier transform
    
    :param qubits: A list of qubit indexes.
    :param coeff: A modifier for the angle used in rotations (-1 for inverse 
                  QFT, 1 for QFT)
    :return: A Quil program to compute the core (inverse) QFT of the qubits.
    """
    
    q = qubits[0]
    qs = qubits[1:]
    if 1 == len(qubits):
        return [H(q)]
    else:
        n = 1 + len(qs)
        cR = []
        for idx, i in enumerate(range(n - 1, 0, -1)):
            q_idx = qs[idx]
            angle = math.pi / 2 ** (n - i)
            cR.append(CPHASE(coeff * angle)(q, q_idx))
        return _core_qft(qs, coeff) + list(reversed(cR)) + [H(q)]


def qft(qubits):
    """
    Generate a program to compute the quantum Fourier transform on
    a set of qubits.

    :param qubits: A list of qubit indexes.
    :return: A Quil program to compute the Fourier transform of the qubits.
    """

    p = pq.Program().inst(_core_qft(qubits, 1))
    return p + bit_reversal(qubits)


def inverse_qft(qubits):
    """
    Generate a program to compute the inverse quantum Fourier transform on
    a set of qubits.

    :param qubits: A list of qubit indexes.
    :return: A Quil program to compute the inverse Fourier transform of the 
             qubits.
    """
    
    qft_result = pq.Program().inst(_core_qft(qubits, -1))
    qft_result += bit_reversal(qubits)
    inverse_qft = pq.Program()
    while len(qft_result.actions) > 0:
        new_inst = qft_result.actions.pop()[1]
        inverse_qft.inst(new_inst)
    return inverse_qft

if __name__ == '__main__':
    print(qft([0, 1, 2, 3]))



import numpy as np

from grove.alpha.arbitrary_state.unitary_operator import unitary_operator, \
    get_bits_needed, fix_norm_and_length


def test_unitary():
    length = 30
    # make an arbitrary complex vector
    v = np.random.uniform(-1, 1, length) \
        + np.random.uniform(-1, 1, length) * 1j

    # normalize and pad with zeros
    v_norm = v / np.linalg.norm(v)
    num_bits = get_bits_needed(length)
    while len(v_norm) < 2 ** num_bits:
        v_norm = np.append(v_norm, 0)

    # generate unitary operator
    U = unitary_operator(v_norm)

    # make sure U|0> = |v>
    zero_state = np.zeros(len(U))
    zero_state[0] = 1
    assert np.allclose(U.dot(zero_state), v_norm)


def test_fix_norm_and_length():
    length = 30
    # make an arbitrary complex vector
    v = np.random.uniform(-1, 1, length) \
        + np.random.uniform(-1, 1, length) * 1j

    new_v = fix_norm_and_length(v)
    assert np.allclose([np.linalg.norm(new_v)], [1])
    assert len(new_v) == 2 ** get_bits_needed(len(v))


def test_get_bits_needed():
    for i, j in zip([1, 2, 5, 8, 9, 14, 27], [1, 1, 3, 3, 4, 4, 5]):
        assert get_bits_needed(i) == j



#!/usr/bin/python
##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from setuptools import setup, find_packages
from grove import __version__

setup(
    name="quantum-grove",
    version=__version__,
    author="Rigetti Computing",
    author_email="softapps@rigetti.com",
    description="A collection of quantum algorithms built using pyQuil and Forest",
    url="https://github.com/rigetticomputing/grove.git",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[
        'numpy',
        'scipy',
        'pyquil >= 1.0.0',
        'mock',
        'networkx',
        'matplotlib'
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'tox',
        'pytest >= 3.0.0',
        'mock'
    ],
    license='LICENSE',
    keywords='quantum quil programming hybrid'
)



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
"""Module containing implementations of oracle Programs.
"""

import numpy as np
import pyquil.quil as pq
from pyquil.gates import X, Z, CNOT

from grove.utils.utility_programs import ControlledProgramBuilder


def basis_selector_oracle(qubits, bitstring):
    """Defines an oracle that selects the ith element of the computational basis.

    Flips the sign of the state :math:`\\vert x\\rangle>`
    if and only if x==bitstring and does nothing otherwise.

    :param qubits: The qubits the oracle is called on. The qubits are assumed to be ordered from
     most significant qubit to least significant qubit.
    :param bitstring: The desired bitstring, given as a string of ones and zeros. e.g. "101"
    :return: A program representing this oracle.
    :rtype: Program
    """
    if len(qubits) != len(bitstring):
        raise ValueError("The bitstring should be the same length as the number of qubits.")
    oracle_prog = pq.Program()

    # In the case of one qubit, we just want to flip the phase of state relative to the other.
    if len(bitstring) == 1:
        oracle_prog.inst(Z(qubits[0]))
        return oracle_prog
    else:
        bitflip_prog = pq.Program()
        for i, qubit in enumerate(qubits):
            if bitstring[i] == '0':
                bitflip_prog.inst(X(qubit))
        oracle_prog += bitflip_prog
        controls = qubits[:-1]
        target = qubits[-1]
        operation = np.array([[1, 0], [0, -1]])
        gate_name = 'Z'
        n_qubit_controlled_z = (ControlledProgramBuilder()
                                .with_controls(controls)
                                .with_target(target)
                                .with_operation(operation)
                                .with_gate_name(gate_name)
                                .build())
        oracle_prog += n_qubit_controlled_z
        oracle_prog += bitflip_prog
    return oracle_prog



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################
import numpy as np
import pyquil.quil as pq
from pyquil.gates import X, H, RZ, CZ

from grove.amplification.amplification import amplification_circuit, diffusion_program
from grove.tests.utils.utils_for_testing import prog_equality

triple_hadamard = pq.Program().inst(H(0)).inst(H(1)).inst(H(2))
cz_gate = pq.Program(CZ(0, 1))
oracle = pq.Program().inst()
qubits = [0, 1, 2]
iters = 2


def test_diffusion_operator():
    """
    Checks that the diffusion operator outputs the correct operation
    """
    created = diffusion_program([0, 1])
    desired = pq.Program()
    for def_gate in created.defined_gates:
        desired.defgate(def_gate.name, def_gate.matrix)
    desired.inst(X(0))
    desired.inst(X(1))
    desired.inst(H(1))
    desired.inst(RZ(-np.pi, 0))
    # There should be only one defined gate -- the CNOT.
    desired.inst((created.defined_gates[0].name, 0, 1))
    desired.inst(RZ(-np.pi, 0))
    desired.inst(H(1))
    desired.inst(X(0))
    desired.inst(X(1))
    assert prog_equality(created, desired)


def test_amplify():
    """
    Test the generic usage of amplify
    """

    # Essentially Grover's to select 011 or 111
    desired = triple_hadamard + cz_gate + triple_hadamard.dagger() + diffusion_program(
        qubits) + triple_hadamard + cz_gate + triple_hadamard.dagger() + diffusion_program(qubits) + triple_hadamard
    created = amplification_circuit(triple_hadamard, cz_gate, qubits, iters)

    prog_equality(desired, created)


def test_amplify_init():
    """
    Test the usage of amplify without init
    """
    # Essentially Grover's to select 011 or 111
    desired = cz_gate + triple_hadamard.dagger() + diffusion_program(
        qubits) + triple_hadamard + cz_gate + triple_hadamard.dagger() + diffusion_program(qubits) + triple_hadamard
    created = amplification_circuit(triple_hadamard, cz_gate, qubits, iters)

    prog_equality(desired, created)



##############################################################################
# Copyright 2016-2017 Rigetti Computing
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
##############################################################################

from pyquil.quil import Program


def isclose(a, b, rel_tol=1e-10, abs_tol=0.0):
    """
    Compares two parameter values.
    :param a: First parameter
    :param b: Second parameter
    :param rel_tol: Relative tolerance
    :param abs_tol: Absolute tolerance
    :return: Boolean telling whether or not the parameters are close enough to be the same
    """
    return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def compare_progs(test, reference):
    """
    Compares two programs gate by gate, param by param.
    :param Program test: Test program
    :param Program reference: Reference program
    """
    tinstr = test.actions
    rinstr = reference.actions
    assert len(tinstr) == len(rinstr)
    for idx in range(len(tinstr)):
        # check each field of the instruction object
        assert tinstr[idx][1].operator_name == rinstr[idx][1].operator_name
        assert len(tinstr[idx][1].parameters) == len(rinstr[idx][1].parameters)
        for pp in range(len(tinstr[idx][1].parameters)):
            cmp_val = isclose(tinstr[idx][1].parameters[pp], rinstr[idx][1].parameters[pp])
            assert cmp_val

        assert len(tinstr[idx][1].arguments) == len(rinstr[idx][1].arguments)
        for aa in range(len(tinstr[idx][1].arguments)):
            assert tinstr[idx][1].arguments[aa] == rinstr[idx][1].arguments[aa]



from grove.alpha.fermion_transforms.fenwick_tree import FenwickTree


def test_sets():
    # tests from 8-qubit example in https://arxiv.org/pdf/1208.5986.pdf
    n_qubits = 8
    tree = FenwickTree(n_qubits)

    parity_dict = {0: set([]),
                   1: set([0]),
                   2: set([1]),
                   3: set([2, 1]),
                   4: set([3]),
                   5: set([4, 3]),
                   6: set([5, 3]),
                   7: set([6, 5, 3])
                  }
    for idx in parity_dict.keys():
        assert set([x.index for x in tree.get_parity_set(idx)]) == \
               parity_dict[idx]

    update_dict = {0: set([1, 3, 7]),
                   1: set([3, 7]),
                   2: set([3, 7]),
                   3: set([7]),
                   4: set([5, 7]),
                   5: set([7]),
                   6: set([7]),
                   7: set([])
                  }
    for idx in update_dict.keys():
        assert set([x.index for x in tree.get_update_set(idx)]) == \
               update_dict[idx]

    flip_dict = {0: set([]),
                 1: set([0]),
                 2: set([]),
                 3: set([2, 1]),
                 4: set([]),
                 5: set([4]),
                 6: set([]),
                 7: set([6, 5, 3])
                }
    for idx in flip_dict.keys():
        assert set([x.index for x in tree.get_children_set(idx)]) == \
               flip_dict[idx]
